/* =============================================================================
 *       Filename:  MyColors.cpp
 *    Description:  RGB and HSV classes
 *
 *                  shamelessy taken and tweaked from:
 *                  http://www.alvyray.com/Papers/CG/hsv2rgb.htm
 *
 *        Created:  10/31/2011 00:12:49
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#include "MyColors.h"
#include <math.h>

RGB_color::RGB_color(float r, float g, float b) : r(r), g(g), b(b) {};

HSV_color::HSV_color(float h, float s, float v) : h(h), s(s), v(v) {};

/* ===  FUNCTION  ==============================================================
 *         Name:  hsv_to_rgb
 *  Description:  converts an hsv value to rgb. used to cycle through color
 *                wheel.
 * ===========================================================================*/
RGB_color hsv_to_rgb( HSV_color HSV ) {
  // H is given on [0, 6] or UNDEFINED. S and V are given on [0, 1].
  // // RGB are each returned on [0, 1].
  float h = HSV.h, s = HSV.s, v = HSV.v, m, n, f;
  int i;
  if (h == -1){
    return RGB_color(v, v, v);
  }
  i = floor(h);
  f = h - i;
  if ( !(i&1) ) f = 1 - f; // if i is even
  m = v * (1 - s);
  n = v * (1 - s * f);
  switch (i) {
    case 6:
    case 0:
      // red to yellow
      return RGB_color(v,n,m);
    case 1:
      // yellow to green
      return RGB_color(n,v,m);
    case 2:
      // green to cyan
      return RGB_color(m,v,n);
    case 3:
      // cyan to blue
      return RGB_color(m,n,v);
    case 4:
      // blue to magenta
      return RGB_color(n,m,v);
    case 5:
      // magenta to red
      return RGB_color(v,m,n);
  }
}
