/*
 * =============================================================================
 *
 *       Filename:  Song.cpp
 *    Description:  Class to store a libsndfile SF_INFO and SNDFILE, as well
 *                  as an array of dots to draw
 *
 *        Created:  12/05/2011 04:44:50
 *
 *         Author:  Derek Tingle (dt), derek.tingle@gmail.com
 *
 * =============================================================================
 */

#include "Definitions.h"
#include "Song.h"


Song::Song(){
  info=new SF_INFO;
  //x=NULL;
  y=NULL;
  z=NULL;
  file=NULL;
}

Song::~Song(){
  if (x)
    delete x;
  if (y)
    delete y;
  if (z)
    delete z;

  delete info;
}

void Song::init(int id){
  db_id=id;
  const char * name;
  std::string sname;
  soci::session sql(soci::sqlite3,DB_PATH);

  try{
  sql << "select name from tracks where id=:id;",
      soci::use(db_id),
      soci::into(sname);

  sname=WAV_DIR+sname;
  name=(const char*)sname.c_str();

  open(name);

  soci::transaction tr(sql);
  sql << "select min(min_mfcc1),max(min_mfcc1),min(min_mfcc2),max(min_mfcc2),min(min_mfcc3),max(min_mfcc3) from r_mfccs;",
      soci::into(min1),
      soci::into(max1),
      soci::into(min2),
      soci::into(max2),
      soci::into(min3),
      soci::into(max3);

  sql << "select count(*) from r_mfccs_data where track_id=:id;",
      soci::use(db_id),
      soci::into(num_points);

  num_points=std::min(num_points,MAX_DOTS);

  x=new std::vector<double>(num_points);
  y=new std::vector<double>(num_points);
  z=new std::vector<double>(num_points);

  sql << "select min_mfcc1,min_mfcc2,min_mfcc3 from r_mfccs join r_mfccs_data on r_mfccs.id=r_mfccs_data.id where track_id=:id limit :max_dots;",
      soci::use(db_id),
      soci::use(num_points),
      soci::into(*x),
      soci::into(*y),
      soci::into(*z);
  //sql << "select count(*) from g
  tr.commit();
  }
  catch( std::exception const &e){
    std::cerr << e.what() << std::endl;
    exit(1);
  }
}

void Song::open(const char * name){
  info->format=0;
  if (DEBUG)
    std::cout << "opening " << name << std::endl;
  file=sf_open(name,SFM_READ,info);
  return;
}

void Song::set_color(RGB_color c){
  color.r=c.r;
  color.g=c.g;
  color.b=c.b;
}

void Song::draw_dots(){
  double x_scaled,y_scaled,z_scaled;
  for (int i=0;i<num_points;++i){
    x_scaled=-1+2*((x->at(i)-min1)/abs(max1-min1));
    y_scaled=-1+2*((y->at(i)-min2)/abs(max2-min2));
    z_scaled=-1+2*((z->at(i)-min3)/abs(max3-min3));
    glPushMatrix();
    glColor4f(color.r,color.g,color.b,(float)DIM_ALPHA-0.03+0.06*(rand()/(float)RAND_MAX));
    glTranslated(x_scaled,y_scaled,z_scaled);
    glutWireCube(DIM_R);
    glPopMatrix();
//    if (i%5000==0)
//      std::cerr << x_scaled << "\t" << y_scaled << "\t" << z_scaled << std::endl;
  }
//    std::cerr << " drawing?" << std::endl;
}

