/* =============================================================================
 *       Filename:  tmp.cpp
 *    Description:  temporary main
 *        Created:  11/15/11 00:43:19
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#include <stdio.h>
#include <string>
#include <exception>
#include <ctime>

#include <boost/thread.hpp>

#include <soci/soci.h>
#include <soci/soci-backend.h>
#include <soci/sqlite3/soci-sqlite3.h>
#include <sndfile.h>
#include <xtract/libxtract.h>

#include "MyAudio.h"
#include "Definitions.h"
#include "extractor/Extractor.h"

using namespace soci;
using namespace std;

// song struct... prolly should be a  class
struct Song{
  SF_INFO *info;
  SNDFILE *file;
};

// function prototypes
int db_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data);
int play_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data);
void prep_songs();


// global for RtAudio buffer
unsigned int g_buffSize = BUFF_SIZE;
// globals for song and data
Song g_in_song;
Song * g_source_songs=NULL;
// stoees number of songs in the tracks table.
int g_num_songs;

// global mutex
boost::mutex g_mutex;
// setup sql
session g_sql;

// extractor
Extractor g_x;
SAMPLE * g_cur_out;
SAMPLE * g_mid_out;

soci::blob * g_bob;

clock_t g_clock;
int g_ticks;

/* ===  FUNCTION  ==============================================================
 *         Name:  db_callback
 *  Description:  callback function passed to RtAudio object. 
 * ===========================================================================*/
int db_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data){
  clock_t cur_clock=clock();
//  debug print something out per callback
//  cerr << ".";
  SAMPLE * outtie = (SAMPLE *) out_buf;
  SAMPLE * innie = (SAMPLE *) in_buf;

  g_mutex.lock();

  g_x.clear_buff();
  // reading from file
  sf_readf_float(g_in_song.file, g_x.buff, (sf_count_t) num_frames);

//  for (int i=0; i<num_frames*NUM_CHANS;++i){
//    outtie[i]=g_x.buff[i];
//  }

  g_x.get_mfccs(g_in_song.info->samplerate);

  int track_id=-1;
  int frame_start=-1;
  Song * cur_song;

  if (USE_BLOBS){
    g_sql << "select track_id, data from r_mfccs join r_mfccs_data on r_mfccs.id=r_mfccs_data.id and min_mfcc1>=:min_m1 and max_mfcc1<=:max_m1 and min_mfcc2>=:min_m2 and max_mfcc2>:max_m2 and min_mfcc3>=:min_m3 and max_mfcc3<=:max_m3 and min_mfcc4>=:min_m4 and max_mfcc4<=:max_m4 and min_mfcc5>=:min_m5 and max_mfcc5<=:max_m5 limit 1;",
        use(g_x.mfccs[0]-2),use(g_x.mfccs[0]+2),
        use(g_x.mfccs[1]-2),use(g_x.mfccs[1]+2),
        use(g_x.mfccs[2]-2),use(g_x.mfccs[2]+2),
        use(g_x.mfccs[3]-2),use(g_x.mfccs[3]+2),
        use(g_x.mfccs[4]-2),use(g_x.mfccs[4]+2),
        into(track_id),into(*g_bob);
    
    if (track_id>0){
//      cerr << "*" << endl;
      g_bob->read((size_t)0,(char *)outtie,(size_t)BUFF_SIZE*NUM_CHANS*sizeof(SAMPLE));
    }
//    else
//     cerr << "." << endl;
  }
  else{
    g_sql << "select track_id, frame_start from r_mfccs join r_mfccs_data on r_mfccs.id=r_mfccs_data.id and min_mfcc1>=:min_m1 and max_mfcc1<=:max_m1 and min_mfcc2>=:min_m2 and max_mfcc2>:max_m2 and min_mfcc3>=:min_m3 and max_mfcc3<=:max_m3 and min_mfcc4>=:min_m4 and max_mfcc4<=:max_m4 and min_mfcc5>=:min_m5 and max_mfcc5<=:max_m5 limit 1;",
        use(g_x.mfccs[0]-2),use(g_x.mfccs[0]+2),
        use(g_x.mfccs[1]-2),use(g_x.mfccs[1]+2),
        use(g_x.mfccs[2]-2),use(g_x.mfccs[2]+2),
        use(g_x.mfccs[3]-2),use(g_x.mfccs[3]+2),
        use(g_x.mfccs[4]-2),use(g_x.mfccs[4]+2),
        into(track_id),into(frame_start);

    if (track_id>0){
//    if (DEBUG)
//      cout << "replacing audio with frame from track " << track_id << endl;
      // gonna change outtie to g_cur_out, then window
//      cerr << "*" << endl;
      cur_song=&g_source_songs[track_id-1];
      sf_seek(cur_song->file,frame_start,SEEK_SET);
      sf_readf_float(cur_song->file,outtie,(sf_count_t) num_frames);
    }
    else{
//      cout << "." << endl;
    }
  }

  g_mutex.unlock();

  g_clock+=clock()-cur_clock;
  ++g_ticks;
  return 0;
}


/* ===  FUNCTION  ==============================================================
 *         Name:  play_callback
 *  Description:  play_callback is a callback that plays a song. nothing fancy.
 * ===========================================================================*/
int play_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data){
//  debug print something out per callback
//  cerr << ".";
  SAMPLE * outtie = (SAMPLE *) out_buf;
  SAMPLE * innie = (SAMPLE *) in_buf;


  // reading from file
  sf_readf_float(g_in_song.file, outtie, (sf_count_t) num_frames);

  return 0;
}


/* ===  FUNCTION  ==============================================================
 *         Name:  prep_songs
 *  Description:  populates g_in_song struct and g_source_songs array of song
 *                structs which store pointers to SF_INFOS and libsnd files
 *
 *                the g_source_song indices correspond to track_id-1
 * ===========================================================================*/
void prep_songs(){
  const char * in_file = "better.wav";
  g_in_song.info = new SF_INFO;
  g_in_song.info->format=0;
  g_in_song.file=sf_open(in_file,SFM_READ,g_in_song.info);
  if (g_in_song.file==NULL){
    cerr << "Error: couldn't open file " << in_file << endl;
    exit(1);
  }
  if (!USE_BLOBS){
    g_sql << "select count(*) from tracks;", into(g_num_songs);

    g_source_songs=new Song[g_num_songs]; 
    string* f_names=new string[g_num_songs];
    const char *cur_file;

    for (int i=0; i < g_num_songs; ++i){
      g_sql << "select name from tracks where id=:id;", use(i+1), into(f_names[i]);
      f_names[i]="data/wav/" + f_names[i];
      cur_file=f_names[i].c_str();
      cerr << "opening " << cur_file << endl;
      g_source_songs[i].info=new SF_INFO;
      g_source_songs[i].info->format=0;
      g_source_songs[i].file=sf_open(cur_file,SFM_READ,g_source_songs[i].info);

      if (g_source_songs[i].file==NULL){
        cerr << "Error: couldn't open file " << f_names[i] << endl;
        exit(1);
      }
    }
  }
}


/* ===  FUNCTION  ==============================================================
 *         Name:  detroy_songs
 *  Description:  delete all of the dynamically allocated song stuff
 * ===========================================================================*/
void destroy_songs(){
  delete g_in_song.info;
  sf_close(g_in_song.file);
  
  if (!USE_BLOBS){
    for (int i=0;i<g_num_songs;++i){
      delete g_source_songs[i].info;
      sf_close(g_source_songs[i].file);
    }
  }
}
/* ===  FUNCTION  ==============================================================
 *         Name:  main
 *  Description:  ?
 * ===========================================================================*/
int main(){
  // open db
  if (USE_BLOBS){
    g_sql.open(soci::sqlite3, "/Users/dt/code/ccrma/256a/project/data/blobs.db");
    g_bob=new soci::blob(g_sql);
  }
  else
    g_sql.open(soci::sqlite3, "/Users/slork/code/ccrma/256a/project/data/ram.db");

  // setup libsndfile
  prep_songs();

  g_cur_out=new SAMPLE[BUFF_SIZE*NUM_CHANS];
  g_mid_out=new SAMPLE[BUFF_SIZE*NUM_CHANS];

  // setup RtAudio
  MyAudio audio(NUM_CHANS, MY_SRATE, &g_buffSize);
  audio.setup(&db_callback, NULL);
//  audio.setup(&play_callback, NULL);
  audio.start();


  // polling block
  char x;
  while(true){
    if (g_ticks>10000){
      cerr << g_clock/g_ticks/(float)CLOCKS_PER_SEC << " seconds per callback"<< endl;
      exit(1);
    }
    usleep(100);
  }

  cerr << "press <enter> to quit" << endl;
  cin.get(x);
  cerr << "exiting" << endl;

  destroy_songs();

  delete g_cur_out;
  delete g_mid_out;

  return 0;
}

