/* =============================================================================
 *       Filename:  tmp.cpp
 *    Description:  temporary main
 *        Created:  11/15/11 00:43:19
 *
 *         Author:  Derek Tingle, derek.tingle@gmail.com
 * ===========================================================================*/

#include <stdio.h>
#include <string>
#include <exception>
#include <ctime>

#include <GLUT/glut.h>
#include <boost/thread.hpp>

#include <soci/soci.h>
#include <soci/soci-backend.h>
#include <soci/sqlite3/soci-sqlite3.h>
#include <sndfile.h>
#include <xtract/libxtract.h>

#include "MyAudio.h"
#include "MyColors.h"
#include "Definitions.h"
#include "extractor/Extractor.h"
#include "Song.h"

using namespace soci;
using namespace std;


/****************************************************************************
 * function prototypes
 * *************************************************************************/
int db_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data);
int play_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data);
void apply_window(SAMPLE* buff,int length);
void get_out_loc(int * out_loc,double * mfccs);
void setup();
void setup_buffs();
void setup_songs();
void init_GL();
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up );
void idleFunc( );
void displayFunc( );
void specialFunc( int key, int x, int y );
void my_view_rotate();

/****************************************************************************
 * GLOBALS
 * *************************************************************************/
// rt_buff_size stores the buffer size that is passed to rt_audio
unsigned int g_rt_buff_size = RT_BUFF_SIZE;
// g_buff_size stores the size of the audio window in the db that the user has 
// selected
int g_buff_size = 8*RT_BUFF_SIZE;
// ratio of user-selected buffer to rt_audio buffer sizes so we know when to 
// extract feats/ overlap add
int g_buff_ratio=g_buff_size/g_rt_buff_size;
// count how many rt_buffers we've ween since extracting features
int g_buff_count=0;

int g_debug_count=-1;

// globals for song and data
Song g_in_song;
Song * g_source_songs=NULL;
// stoees number of songs in the tracks table.
int g_num_songs;
// flag that is set to true when song finishes
bool g_song_done=false;
// flag to tell audio callback to play song
bool g_play=false;

// global mutex
boost::mutex g_mutex;
// setup sql
session g_sql;

// extractors
Extractor *g_cur;
Extractor *g_mid;

// buffer to store outputs
// a brief note about naming, mid refers to the user-defined windows that fall
// between the middle of the previous and the middle of the current output
// buffers. current refers to buffers that sync with the output buffers. this
// distinction becomes hazy when the buffer_ratio isn't one
SAMPLE *g_cur_out;
SAMPLE *g_mid_out;

clock_t g_clock;
int g_ticks;

const char *g_in_name;
double g_max1,g_min1,g_max2,g_min2,g_max3,g_min3;

// ==========================
// GL globals
// ==========================
// Camera control global variables
pt3d g_look_from( 0, 0, 1);
pt3d g_look_to( 0, 0, 0 );
pt3d g_head_up( 0, 1, 0 );
// amt to change the camera by
GLfloat g_delta=5;
// instead of rotating the camera, we'll just rotate the model
pt3d g_view_rotate(0,0,0);
GLsizei g_width = 800; 
GLsizei g_height = 600;
bool g_is_full = false;

// struct defined in definitions
lit_dot g_brights[LIT_DOTS];
int g_bright_pos=0;

SAMPLE * my_write_buff;
// 
/* ===  FUNCTION  ==============================================================
 *         Name:  db_callback
 *  Description:  callback function passed to RtAudio object. 
 * ===========================================================================*/
int write_callback(SAMPLE * outtie, SNDFILE *write_file){
//  debug print something out per callback
//  cerr << ".";

  // read into g_cur.buff
  if (sf_readf_float(g_in_song.file, &g_cur->buff[g_buff_count*RT_BUFF_SIZE*NUM_CHANS], (sf_count_t) RT_BUFF_SIZE)==0)
    g_song_done=true;

  // since the buffer that we actually want to use to read and resynthesize may
  // be much larger than the RtAudio buffer, we'll just wait until enough
  // buffers have passed before we read and do the calculations
  // file 
  if (++g_buff_count==g_buff_ratio){
    // reset counter
    g_buff_count=0;

    Song * cur_song;
    int out_loc[2];
    int frames_read;

    // copy first half of current buffer into second half of mid_buffer
    // the first half of mid buffer should have the second half of cur_buff 
    // from last callback
    for (int i=0;i<NUM_CHANS*g_buff_size/2;++i){
      g_mid->buff[i+NUM_CHANS*g_buff_size/2]=g_cur->buff[i];
    }

    g_mid->get_mfccs(g_in_song.info->samplerate);
    get_out_loc(out_loc,g_mid->mfccs);
    memset(g_mid_out,0,g_buff_size*NUM_CHANS*sizeof(SAMPLE));

    // find matching frame in db and read it into g_mid_out
    if (out_loc[0]>0){
      cur_song=&g_source_songs[out_loc[0]-1];
      sf_seek(cur_song->file,out_loc[1],SEEK_SET);
      sf_readf_float(cur_song->file,g_mid_out,(sf_count_t) g_buff_size);
      apply_window(g_mid_out,g_buff_size*NUM_CHANS);

    }
    else{
      cerr << "." ;
      for (int i=0;i<NUM_CHANS*g_buff_size;++i){
        g_mid_out[i]=0;
      }
    }

    // copy the second half of the previous cur_out into outtie (the previous
    // cur_out was already windowed
    for (int i=0;i<NUM_CHANS*g_buff_size/2;i++){
      g_mid_out[i]+=g_cur_out[NUM_CHANS*g_buff_size/2+i];
    }

    memset(g_cur_out,0,NUM_CHANS*g_buff_size);

    // now find the best match in the db for the current input frame
    g_cur->get_mfccs(g_in_song.info->samplerate);
    get_out_loc(out_loc,g_cur->mfccs);

    if (out_loc[0]>0){
      cur_song=&g_source_songs[out_loc[0]-1];
      sf_seek(cur_song->file,out_loc[1],SEEK_SET);
      sf_readf_float(cur_song->file,g_cur_out,(sf_count_t) g_buff_size);
      apply_window(g_cur_out,g_buff_size*NUM_CHANS);

    }
    else{
      //debug
      cerr << "." ;
      for (int i=0;i<NUM_CHANS*g_buff_size;++i){
        g_cur_out[i]=0;
      }
    }

    g_mid->clear_buff();

    // copy the first half of this cur_out into the second half of outtie
    // AND copy the second half of this cur_buff into the first half of mid_buff
    // (for next callback)
    for (int i=0;i<NUM_CHANS*g_buff_size/2;i++){
      g_mid_out[i+NUM_CHANS*g_buff_size/2]+=g_cur_out[i];
      g_mid->buff[i]=g_cur->buff[i+NUM_CHANS*g_buff_size/2];
    }

    g_cur->clear_buff();

    for (int i=0;i<NUM_CHANS*g_buff_size;++i){
      if (fabs(g_mid_out[i])>1){
        cerr << i << endl;
      } 
      my_write_buff[i]=g_mid_out[i];
    }
    sf_writef_float(write_file,my_write_buff,NUM_CHANS*g_buff_size);
    //cout << clock()-cur_clock << "\t" <<  g_dot_start << endl;
  }

  // if the current RtAudio buffer isn't halfway through the desired buffer len

  else{
  }
  // we wanna do this anyway:
 // for(int i=0;i<RT_BUFF_SIZE*NUM_CHANS;++i){
 //   my_write_buff[i]=g_mid_out[(RT_BUFF_SIZE*NUM_CHANS*g_buff_count)+i];
 // }
  //sf_writef_float(write_file,my_write_buff,RT_BUFF_SIZE);

  //++g_debug_count;
  return 0;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  apply_window
 *  Description:  apply a hanning window to the passed in buffer
 * ===========================================================================*/
void apply_window(SAMPLE* buff,int length){
  double pi, phase=0, delta;

  delta = 2 * PI / (double) length;

  for(int i = 0; i < length; i++ ){
    buff[i] = buff[i]*(float)(0.5 * (1.0 - cos(phase)));
    phase += delta;
  }

  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  get_buff_loc
 *  Description:  query the database for the closest samples to a given vector
 *                of MFCCs. returns a track_id and a frame_start
 * ===========================================================================*/
void get_out_loc(int * out_loc,double * mfccs){
  memset(out_loc,-1,sizeof(int)*2);

  try{
    g_sql << "select track_id, frame_start from r_mfccs join r_mfccs_data on r_mfccs.id=r_mfccs_data.id where buff_size=:buff_size and min_mfcc1>=:min_m1 and max_mfcc1<=:max_m1 and min_mfcc2>=:min_m2 and max_mfcc2<:max_m2 and min_mfcc3>=:min_m3 and max_mfcc3<=:max_m3 and min_mfcc4>=:min_m4 and max_mfcc4<=:max_m4 and min_mfcc5>=:min_m5 and max_mfcc5<=:max_m5 order by random() limit 1;",
        use(g_buff_size),
        use(mfccs[0]-10),use(mfccs[0]+10),
        use(mfccs[1]-10),use(mfccs[1]+10),
        use(mfccs[2]-10),use(mfccs[2]+10),
        use(mfccs[3]-10),use(mfccs[3]+10),
        use(mfccs[4]-10),use(mfccs[4]+10),
        into(out_loc[0]),into(out_loc[1]);
  }
  catch( exception const & e){
    cerr << e.what() << endl;
    exit(1);
  }

  return;
}

/* ===  FUNCTION  ==============================================================
 *         Name:  play_callback
 *  Description:  play_callback is a callback that plays a song. nothing fancy.
 * ===========================================================================*/
int play_callback(void * out_buf, void * in_buf, unsigned int num_frames,
              double stream_time, RtAudioStreamStatus status, void * data){
//  debug print something out per callback
//  cerr << ".";
  SAMPLE * outtie = (SAMPLE *) out_buf;
  SAMPLE * innie = (SAMPLE *) in_buf;


  // reading from file
  sf_readf_float(g_in_song.file, outtie, (sf_count_t) num_frames);

  return 0;
}


/* ===  FUNCTION  ==============================================================
 *         Name:  setup
 *  Description:  initializes useful globals and sets up GL
 *
 *                populates g_in_song struct and g_source_songs array of song
 *                structs which store pointers to SF_INFOS and libsnd files
 *
 *                the g_source_song indices correspond to track_id-1
 * ===========================================================================*/
void setup(){

  setup_buffs();
  setup_songs();
}

/* ===  FUNCTION  ==============================================================
 *         Name:  setup_buffs
 *  Description:  b/c i wanna be able to change the buffers on the fly 
 * ===========================================================================*/
void setup_buffs(){
  g_buff_ratio=g_buff_size/RT_BUFF_SIZE;

  g_cur_out=new SAMPLE[g_buff_size*NUM_CHANS];
  memset(g_cur_out,0,NUM_CHANS*g_buff_size*(sizeof(SAMPLE)));

  g_mid_out=new SAMPLE[g_buff_size*NUM_CHANS];
  memset(g_mid_out,0,NUM_CHANS*g_buff_size*(sizeof(SAMPLE)));

  g_cur=new Extractor(g_buff_size);
  g_mid=new Extractor(g_buff_size);
}


/* ===  FUNCTION  ==============================================================
 *         Name:  setup_songs
 *  Description: populates g_in_song struct and g_source_songs array of song
 *               structs which store pointers to SF_INFOS and libsnd files
 *
 *                the g_source_song indices correspond to track_id-1
 * ===========================================================================*/
void setup_songs(){
  // setup input song and source songs
  //const char * in_file = "better.wav";
  g_in_song.info = new SF_INFO;
  g_in_song.info->format=0;
  g_in_song.file=sf_open(g_in_name,SFM_READ,g_in_song.info);
  if (g_in_song.file==NULL){
    cerr << "Error: couldn't open file " << g_in_name << endl;
    exit(1);
  }

  g_sql << "select count(*) from tracks;", into(g_num_songs);

  g_source_songs=new Song[g_num_songs]; 
  string* f_names=new string[g_num_songs];
  const char *cur_file;

  float color_inc=(float)6/g_num_songs;
  for (int i=0; i < g_num_songs; ++i){
    g_source_songs[i].init(i+1);
    g_source_songs[i].set_color(hsv_to_rgb(HSV_color(color_inc*i,1.0f,1.0f)));

    if (g_source_songs[i].file==NULL){
      cerr << "Error: couldn't open file " << f_names[i] << endl;
      exit(1);
    }
  }
}


//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
  switch( key ){
    case 'Q':
    case 'q':
  //    for (int i=0;i<NUM_DOTS;++i){
  //      cerr << g_dots[i].x << g_dots[i].y << g_dots[i].z << "\t";
  //    }
      cout << g_clock/g_ticks/(float)CLOCKS_PER_SEC << " seconds per callback"<< endl;
      exit(0);
      break;
    case 'f':
    case 'F':
      // fullscreen
      if (g_is_full){
        glutReshapeWindow(800,500);
      }
      else
        glutFullScreen();

      g_is_full=!g_is_full;
      break;
    // pause or play
    case 'p':
    case 'P':
      cerr << "pausing or playing" << endl;
      g_play=!g_play;
  }
  glutPostRedisplay( );
}

//----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles special function keys
//----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
  if( key == GLUT_KEY_LEFT)
  {
    g_view_rotate.x += g_delta;
  }
  if( key == GLUT_KEY_RIGHT)
  {
    g_view_rotate.x -= g_delta;
  }
  if( key == GLUT_KEY_DOWN)
  {
    g_look_from.z += (g_delta/100);
  }
  if( key == GLUT_KEY_UP)
  {
    g_look_from.z -= (g_delta/500);
  }
  if( key == GLUT_KEY_PAGE_UP)
  {
    std::cerr << "PageUp arrow";
  }
  if( key == GLUT_KEY_PAGE_DOWN)
  {
    std::cerr << "PageDown arrow";
  }

  glutPostRedisplay( );
}
//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
  // save the new window size
  g_width = (GLsizei)w; 
  g_height = (GLsizei)h;
  // map the view port to the client area
  glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
  // set the matrix mode to project
  glMatrixMode( GL_PROJECTION );
  // load the identity matrix
  glLoadIdentity( );
  // create the viewing frustum
  gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 0.1, 300.0 );
  // set the matrix mode to modelview
  glMatrixMode( GL_MODELVIEW );
  // load the identity matrix
  glLoadIdentity( );
  // position the view point
  changeLookAt( g_look_from,g_look_to, g_head_up );
}

void init_GL(){

  // =================
  // = OpenGL & GLUT =
  // =================

  // set the GL clear color - use when the color buffer is cleared
  glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    
  // set the shading model to 'smooth'
  glShadeModel( GL_SMOOTH );
  // enable depth
  glEnable( GL_DEPTH_TEST );
  // set the front faces of polygons
  glFrontFace( GL_CCW );
  // set fill mode
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

  // Enable transparency
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
  // seed random number generator
  srand( time(NULL) );


}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
  // render the scene
  glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
  // clear the color and depth buffers
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // save the current transformation
  glPushMatrix();
    

    //glLineWidth( 1.0f );  

  
    glColor4f(1, 1, 0, .3);


    changeLookAt( g_look_from, g_look_to, g_head_up );
    my_view_rotate();
    for (int i=0;i<g_num_songs;++i){
      g_source_songs[i].draw_dots();
    }

    g_mutex.lock();
    for (int i=0;i<LIT_DOTS;++i){
      glPushMatrix();
      glColor4f(g_brights[i].color.r,g_brights[i].color.g,g_brights[i].color.b,g_brights[i].pos.a);
      g_brights[i].pos.a-=(1-DIM_ALPHA)/5;
      glTranslatef(g_brights[i].pos.x,g_brights[i].pos.y,g_brights[i].pos.z);
      glutWireCube(BRIGHT_R);
      glPopMatrix();
    }
    g_mutex.unlock();
//    for (int i=0;i<NUM_DOTS;++i){
//      if ((g_dots[i].x!=0) || (g_dots[i].y!=0) ||(g_dots[i].z!=0)){
//        glPushMatrix();
//        glTranslatef(g_dots[i].x,g_dots[i].y,g_dots[i].z);
////        glutSolidSphere(.005,10,10);
//        glutWireCube(.005);
//        
//        glPopMatrix();
//      }
//    }
//    g_mutex.unlock();
    // set the view point (camera)
//    changeLookAt( g_look_from, g_look_to, g_head_up );

  // restore state
  glPopMatrix();


  // flush!
  glFlush( );
  // swap the double buffer
  glutSwapBuffers( );
}

//-----------------------------------------------------------------------------
// name: changeLookAt()
// desc: changes the point of view
//-----------------------------------------------------------------------------
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up )
{
  gluLookAt(  look_from.x, look_from.y, look_from.z, 
        look_to.x, look_to.y, look_to.z, 
        head_up.x, head_up.y, head_up.z );
}

/* ===  FUNCTION  ==============================================================
 *         Name:  my_view_rotate
 *  Description:  rotates the model (so it looks like the camera's rotating
 * ===========================================================================*/
void my_view_rotate(){
  if (g_view_rotate.x!=0){
    g_view_rotate.x = fmod(g_view_rotate.x,360);
    glRotatef(g_view_rotate.x,0,1,0);
//    g_look_from.x = ::sin(2*PI*g_view_rotate.x/360);
  }
}

//-----------------------------------------------------------------------------
// name: drawAxis()
// desc: draw cartesian axis (x,y,z) using (r,g,b) respectively
//-----------------------------------------------------------------------------
void drawAxis()
{
  glPushMatrix();
    
  glBegin( GL_LINES );
  
    // x axis
    glColor4f( 1, 0, 0, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 1, 0, 0 );

    // y axis
    glColor4f( 0, 1, 0, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 0, 1, 0 );

    // z axis
    glColor4f( 0, 0, 1, 1 );
    glVertex3f( 0, 0, 0 );
    glVertex3f( 0, 0, 1 );

  glEnd();
  glPopMatrix();
}

/* ===  FUNCTION  ==============================================================
 *         Name:  detroy_songs
 *  Description:  delete all of the dynamically allocated song stuff
 * ===========================================================================*/
void destroy_songs(){
  delete g_in_song.info;
  sf_close(g_in_song.file);
  
  for (int i=0;i<g_num_songs;++i){
    delete g_source_songs[i].info;
    sf_close(g_source_songs[i].file);
  }
}
/* ===  FUNCTION  ==============================================================
 *         Name:  main
 *  Description:  ?
 * ===========================================================================*/
int main(int argc, char *argv[]){
  g_in_name="/Users/slork/code/ccrma/256a/project/stronger.wav";

  my_write_buff=new SAMPLE[g_buff_size*NUM_CHANS];

  // open db
  g_sql.open(soci::sqlite3, DB_PATH);

  // setup some stuff
  setup();

  SF_INFO *write_info=new SF_INFO;
  write_info->format=SF_FORMAT_WAV|SF_FORMAT_PCM_16;
  write_info->samplerate=44100;
  write_info->channels=2;
  const char *out_name="mash16.wav";

  SNDFILE *write_file=sf_open(out_name,SFM_WRITE,write_info);

  while (!g_song_done){
    memset(my_write_buff,0,sizeof(SAMPLE)*RT_BUFF_SIZE*NUM_CHANS);
    write_callback(my_write_buff,write_file);
  }

  sf_close(write_file);

  // polling block
//  char x;
//  while(true){
//    if ((g_song_done) || (g_ticks>10000)){
//      cerr << g_clock/g_ticks/(float)CLOCKS_PER_SEC << " seconds per callback"<< endl;
//      break;
//    }
//
//    usleep(100);
//  }


  return 0;
}

